import torch
import torch.nn as nn
from torch.nn.parameter import Parameter


class BatchNorm(nn.Module):
    @classmethod
    def from_onnx(cls, mod):
        bn = nn.BatchNorm2d(mod.inputs[1].shape[0], eps=mod.attrs["epsilon"])
        bn.scale = Parameter(torch.from_numpy(mod.inputs[1].values))
        bn.bias = Parameter(torch.from_numpy(mod.inputs[2].values))
        bn.running_mean = Parameter(
            torch.from_numpy(mod.inputs[3].values), requires_grad=False
        )
        bn.running_var = Parameter(
            torch.from_numpy(mod.inputs[4].values), requires_grad=False
        )

        return bn



import torch
import torch.nn as nn
from onnx_graphsurgeon import Constant


class Arithmetic(nn.Module):
    @classmethod
    def from_onnx(cls, mod):  # arithmetic can be different
        def get_input_node(input):
            if isinstance(input, Constant):
                return torch.nn.Parameter(
                    torch.from_numpy(input.values), requires_grad=False
                )
            else:
                return input

        inputs = []
        for input in mod.inputs:
            if len(input.inputs) == 0 and not isinstance(
                input, Constant
            ):  # input is real input
                inputs.append(input)
            elif isinstance(input, Constant):
                inputs.append(get_input_node(input))
            for feed in input.inputs:
                if (
                    feed.op == "Split"
                ):  # for split node, we need to get the output of split node
                    inputs.append(input)
                else:
                    inputs.append(get_input_node(feed))

        return tuple(inputs)



from typing import Union

import onnx


def convert(
    model: Union[str, onnx.ModelProto],
    output_model: str = None,
    check: bool = False,
):
    from o2t.core.parser import OnnxPytorchParser

    onnx2torch = OnnxPytorchParser(model)
    onnx2torch.convert()

    if check:
        onnx2torch.check()

    if not output_model:
        return onnx2torch.model
    else:
        onnx2torch.save(output_model)


def main():
    import argparse

    from loguru import logger

    import o2t

    parser = argparse.ArgumentParser(
        description="o2t: Onnx to Pytorch Converter",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("input_model", help="input onnx model")
    parser.add_argument(
        "output_model", nargs="?", default=None, help="output onnx model"
    )

    parser.add_argument("--check", action="store_true", help="enable model check")
    parser.add_argument("-v", "--version", action="version", version=o2t.__version__)

    args, unknown = parser.parse_known_args()

    if unknown:
        logger.error(f"unrecognized options: {unknown}")
        return 1

    convert(
        args.input_model,
        args.output_model,
        args.check,
    )

    return 0



import _operator
import re

import numpy as np

import onnx
import onnx_graphsurgeon as gs
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.fx import Graph, GraphModule
from .pytorch_layers import *
from ..utils.utils import (
    gen_onnxruntime_input_data,
    numpy_dtype_to_torch,
    onnxruntime_inference,
)


class OnnxPytorchParser:
    def __init__(self, model, block_info=None):
        super(OnnxPytorchParser, self).__init__()
        self.model = model
        self.block_info = block_info
        if isinstance(model, str):
            self.onnx_model = onnx.load(model)
        else:
            self.onnx_model = model

        self.graph = gs.import_onnx(self.onnx_model)
        self.graph.fold_constants().cleanup().toposort()
        self.pytorch_graph = Graph()
        self.pytorch_graph_module = GraphModule(torch.nn.Module(), self.pytorch_graph)
        self.env = {}
        self._illegal_char_regex = re.compile("[^0-9a-zA-Z_]+")

    def convert(self):
        self.gen_pytorch_graph_module()

    def create_arg(self, a):
        if isinstance(a, torch.nn.Parameter):
            for n, p in self.pytorch_graph_module.named_parameters():
                if a is p:
                    return self.create_node("get_attr", n, (), {})
        elif isinstance(a, torch.Tensor):
            for n_, p_ in self.pytorch_graph_module.named_buffers():
                if a is p_:
                    return self.create_node("get_attr", n_, (), {})
        elif isinstance(a, torch.nn.Module):
            for n_, p_ in self.pytorch_graph_module.named_modules():
                if a is p_:
                    return self.create_node("get_attr", n_, (), {})

        if isinstance(a, tuple) and hasattr(a, "_fields"):
            args = tuple(self.create_arg(elem) for elem in a)
            return self.create_node("call_function", a.__class__, args, {})

        qualname = None
        if isinstance(a, (torch.Tensor)):
            if not qualname:
                i = 0
                while True:
                    qualname = f"_tensor_constant{i}"
                    if not hasattr(self.pytorch_graph_module, qualname):
                        break
                    i += 1
                setattr(self.pytorch_graph_module, qualname, a)

            return self.pytorch_graph.create_node("get_attr", qualname, (), {})

    def process_inputs(self, inputs):
        inputs = list(inputs)
        for idx in range(len(inputs)):
            input = self.create_arg(inputs[idx])
            if input:
                inputs[idx] = input
            else:
                inputs[idx] = self.env[inputs[idx].name]

        inputs = tuple(inputs)

        return inputs

    def get_node_users(self, node):
        users = []
        for output in node.outputs:  # output is a Variable
            for user in output.outputs:  # user is a Node
                users.append(user)
        return users

    def get_node_feeds(self, node):
        feeds = []
        for input in node.inputs:  # input is a Variable
            for feed in input.inputs:  # user is a Node
                if (
                    feed.op == "Split"
                ):  # for split node, we need to get the output of split node
                    feeds.append(input)
                else:
                    feeds.append(feed)
        return feeds

    def find_block_id(self, node_name, block_info):
        if block_info is None:
            return None

        for block_id, block_data in block_info.items():
            if node_name in block_data["nodes"]:
                return block_id
        # Return None if the node is not found in any block
        return None

    def get_value_by_key_or_index(self, node, key, index, default=None):
        if key in node.attrs:
            return node.attrs[key]
        elif index < len(node.inputs):
            if isinstance(node.inputs[index], gs.Constant):
                return node.inputs[index].values
            else:
                return default
        else:
            return default

    def gen_pytorch_graph_module(self):
        for input in self.graph.inputs:
            node = self.pytorch_graph.placeholder(
                self._illegal_char_regex.sub("_", input.name)
            )
            self.env[input.name] = node

        for onnx_node in self.graph.nodes:
            node_name = onnx_node.name
            target_name = node_name
            block_id = self.find_block_id(node_name, self.block_info)
            if block_id is not None:
                target_name = f"{block_id}.{node_name}"
            node_feeds = self.get_node_feeds(onnx_node)
            if len(node_feeds) == 0:
                node_feeds = self.graph.inputs[0]
            elif len(node_feeds) == 1:
                node_feeds = node_feeds[0]

            if onnx_node.op == "Conv":
                module = Conv.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "LayerNormalization":
                module = LayerNorm.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Relu":
                module = nn.ReLU()
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Clip":
                module = nn.ReLU6()
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Add":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.add,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Sub":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.sub,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Div":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.div,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Mul":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.mul,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "MatMul":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.matmul,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Gelu":
                node = self.pytorch_graph.create_node(
                    "call_function",
                    F.gelu,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "LogSoftmax":
                axis = onnx_node.attrs.get("axis", None)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    F.log_softmax,
                    (self.env[node_feeds.name], axis),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "GlobalAveragePool":
                module = Pool.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "MaxPool":
                module = Pool.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "AveragePool":
                module = Pool.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Flatten":
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.flatten,
                    (self.env[node_feeds.name],),
                    {"start_dim": onnx_node.attrs["axis"]},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Concat":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.cat,
                    (inputs,),
                    {"dim": onnx_node.attrs["axis"]},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Reshape":
                if isinstance(onnx_node.inputs[1], gs.Constant):
                    node = self.pytorch_graph.create_node(
                        "call_method",
                        "reshape",
                        (
                            self.env[node_feeds.name],
                            onnx_node.inputs[1].values.tolist(),
                        ),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
                else:
                    cpu_node_name = node_name + "_cpu"
                    node = self.pytorch_graph.create_node(
                        "call_method",
                        "cpu",
                        (self.env[node_feeds[1].name],),
                        {},
                        cpu_node_name,
                    )
                    self.env[cpu_node_name] = node
                    numpy_node_name = node_name + "_numpy"
                    node = self.pytorch_graph.create_node(
                        "call_method",
                        "numpy",
                        (self.env[cpu_node_name],),
                        {},
                        numpy_node_name,
                    )
                    self.env[numpy_node_name] = node
                    to_list_node_name = node_name + "_tolist"
                    node = self.pytorch_graph.create_node(
                        "call_method",
                        "tolist",
                        (self.env[numpy_node_name],),
                        {},
                        to_list_node_name,
                    )
                    self.env[to_list_node_name] = node
                    module = Reshape.from_onnx()
                    self.pytorch_graph_module.add_submodule(target_name, module)
                    node = self.pytorch_graph.create_node(
                        "call_module",
                        target_name,
                        (self.env[node_feeds[0].name], self.env[to_list_node_name]),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
            elif onnx_node.op == "Transpose":
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.permute,
                    (
                        self.env[node_feeds.name],
                        onnx_node.attrs["perm"],
                    ),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Split":
                chunk = self.get_value_by_key_or_index(
                    onnx_node, "split", 1, len(onnx_node.outputs)
                )
                if len(onnx_node.inputs) > 1:
                    if isinstance(onnx_node.inputs[1], gs.Constant):
                        chunk = onnx_node.inputs[1].values
                    else:
                        chunk = self.env[node_feeds[1].name]
                    func = torch.split
                else:
                    chunk = len(onnx_node.outputs)
                    func = torch.tensor_split

                if isinstance(chunk, np.ndarray):
                    chunk = chunk.tolist()
                dim = self.get_value_by_key_or_index(onnx_node, "axis", 2, 0)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    func,
                    (
                        self.env[node_feeds.name],
                        chunk,
                        dim,
                    ),
                    {},
                    node_name,
                )
                self.env[node_name] = node
                for i, output in enumerate(onnx_node.outputs):
                    node = self.pytorch_graph.create_node(
                        "call_function",
                        _operator.getitem,
                        (
                            self.env[node_name],
                            i,
                        ),
                        {},
                        output.name,
                    )
                    self.env[output.name] = node
            elif onnx_node.op == "Slice":
                if isinstance(onnx_node.inputs[0], gs.Constant):
                    node_feeds = torch.nn.Parameter(
                        torch.from_numpy(onnx_node.inputs[0].values),
                        requires_grad=False,
                    )
                    node_feeds = self.process_inputs([node_feeds])[0]
                elif isinstance(node_feeds, list):
                    node_feeds = node_feeds[0]
                    node_feeds = self.env[node_feeds.name]
                else:
                    node_feeds = self.env[node_feeds.name]
                inputs = Slice.from_onnx(onnx_node, self.env)
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    _operator.getitem,
                    (
                        node_feeds,
                        inputs,
                    ),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Gemm":
                module = Linear.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "BatchNormalization":
                module = BatchNorm.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Softmax":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    F.softmax,
                    (self.env[node_feeds.name],),
                    {"dim": -1},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Sigmoid":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    F.sigmoid,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "HardSwish":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    F.hardswish,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "LeakyRelu":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    F.leaky_relu,
                    (self.env[node_feeds.name], onnx_node.attrs["alpha"]),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Resize":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    F.interpolate,
                    (self.env[node_feeds.name],),
                    {
                        "scale_factor": onnx_node.inputs[2].values.tolist()[2:],
                        "mode": onnx_node.attrs["mode"],
                    },
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Pow":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.pow,
                    (self.env[node_feeds.name], float(onnx_node.inputs[1].values)),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Sqrt":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.sqrt,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Erf":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.sqrt,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Squeeze":
                dim = self.get_value_by_key_or_index(onnx_node, "axes", 1, None)
                if isinstance(dim, np.ndarray):
                    dim = dim.tolist()
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.squeeze,
                    (self.env[node_feeds.name], dim),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Unsqueeze":
                axes = onnx_node.attrs.get("axes")[0]
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.unsqueeze,
                    (self.env[node_feeds.name], int(axes)),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Neg":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.neg,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "ConstantOfShape":
                size_name = node_name + "_size"
                module = Size.from_onnx()
                self.pytorch_graph_module.add_submodule(size_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    size_name,
                    (self.env[node_feeds.name],),
                    {},
                    size_name,
                )
                self.env[size_name] = node
                module = Full.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[size_name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "ReduceMean":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_method",
                    "mean",
                    (self.env[node_feeds.name],),
                    {
                        "dim": onnx_node.attrs["axes"],
                        "keepdim": bool(onnx_node.attrs.get("keepdims", 1)),
                    },
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Cast":
                torch_dtype = numpy_dtype_to_torch(
                    onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[onnx_node.attrs["to"]]
                )
                node = self.pytorch_graph_module.graph.create_node(
                    "call_method",
                    "to",
                    (self.env[node_feeds.name], torch_dtype),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "ReduceSum":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.sum,
                    (self.env[node_feeds.name],),
                    {
                        "dim": onnx_node.attrs["axes"],
                        "keepdim": bool(onnx_node.attrs.get("keepdims", 1)),
                    },
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "ReduceMax":
                if onnx_node.attrs.get("axes"):
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        torch.max,
                        (self.env[node_feeds.name],),
                        {
                            "dim": onnx_node.attrs.get("axes"),
                            "keepdim": bool(onnx_node.attrs.get("keepdims", 1)),
                        },
                        node_name,
                    )
                    self.env[node_name] = node
                else:
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        torch.max,
                        (self.env[node_feeds.name],),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
            elif onnx_node.op == "Shape":
                shape_node_name = node_name + "_shape"
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    getattr,
                    (self.env[node_feeds.name], "shape"),
                    {},
                    shape_node_name,
                )
                self.env[shape_node_name] = node
                module = Tensor.from_onnx()
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[shape_node_name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Range":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                module = Arange.from_onnx()
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Equal":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.eq,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "LessOrEqual":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.less_equal,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "GreaterOrEqual":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.greater_equal,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "And":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.logical_and,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Not":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.logical_not,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Where":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.where,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Expand":
                inputs = Arithmetic.from_onnx(onnx_node)
                inputs = self.process_inputs(inputs)

                size_name = node_name + "_size"
                module = Size.from_onnx()
                self.pytorch_graph_module.add_submodule(size_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    size_name,
                    (self.env[node_feeds[1].name],),
                    {},
                    size_name,
                )
                self.env[size_name] = node

                ones_node_name = node_name + "_ones"
                module = Ones.from_onnx()
                self.pytorch_graph_module.add_submodule(ones_node_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    ones_node_name,
                    (self.env[size_name],),
                    {},
                    ones_node_name,
                )
                self.env[ones_node_name] = node

                node = self.pytorch_graph.create_node(
                    "call_function",
                    _operator.mul,
                    (self.env[node_feeds[0].name], self.env[ones_node_name]),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Gather":
                if isinstance(onnx_node.inputs[0], gs.Constant):
                    module = Embedding.from_onnx(onnx_node)
                    self.pytorch_graph_module.add_submodule(target_name, module)
                    node = self.pytorch_graph.create_node(
                        "call_module",
                        target_name,
                        (self.env[node_feeds.name],),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
                elif all(isinstance(input, gs.Variable) for input in onnx_node.inputs):
                    axis = onnx_node.attrs.get("axis", 0)
                    index = self.env[node_feeds[1].name]
                    if axis == 0:
                        index_all = index
                    else:
                        index_all = [slice(None, None, None)] * (axis)
                        index_all.append(index)
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        _operator.getitem,
                        (
                            self.env[node_feeds[0].name],
                            index_all,
                        ),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
                else:
                    axis = onnx_node.attrs.get("axis", 0)
                    index = torch.nn.Parameter(
                        torch.from_numpy(onnx_node.inputs[1].values),
                        requires_grad=False,
                    )
                    index = self.process_inputs([index])[0]
                    if axis == 0:
                        index_all = index
                    else:
                        index_all = [slice(None, None, None)] * (axis)
                        index_all.append(index)
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        _operator.getitem,
                        (
                            self.env[node_feeds.name],
                            index_all,
                        ),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
            elif onnx_node.op == "QuantizeLinear":
                dequant_node = onnx_node.o(0)
                assert dequant_node.op == "DequantizeLinear"

                module = Observer(
                    float(onnx_node.inputs[1].values), float(onnx_node.inputs[2].values)
                )
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds.name],),
                    {},
                    node_name,
                )
                self.env[dequant_node.outputs[0].name] = node
            elif onnx_node.op == "DequantizeLinear":
                pass
            else:
                raise NotImplementedError(
                    "Operator {} is not supported.".format(onnx_node.op)
                )

        if len(self.graph.outputs) == 1:
            if self.graph.outputs[0].inputs[0].op == "Split":
                graph_output = self.env[self.graph.outputs[0].name]
            else:
                graph_output = self.env[self.graph.outputs[0].inputs[0].name]

            node = self.pytorch_graph.output(graph_output)
        else:
            graph_output = []
            for output in self.graph.outputs:
                if output.inputs[0].op == "Split":
                    graph_output.append(self.env[output.name])
                else:
                    graph_output.append(self.env[output.inputs[0].name])

            node = self.pytorch_graph.output(graph_output)

        self.pytorch_graph_module.graph.lint()
        self.pytorch_graph_module.recompile()

    def save(self, output_model):
        torch.save(self.pytorch_graph_module, output_model)

    def check(self):
        input_data_dict = gen_onnxruntime_input_data(self.onnx_model)
        onnx_output_dict = onnxruntime_inference(self.onnx_model, input_data_dict)
        torch_dict = {
            self._illegal_char_regex.sub("_", k): torch.from_numpy(v)
            for k, v in input_data_dict.items()
        }
        with torch.no_grad():
            self.pytorch_graph_module.eval()
            torch_output = self.pytorch_graph_module(**torch_dict)

        if isinstance(torch_output, torch.Tensor):
            torch_output = [torch_output]

        onnx_output = list(onnx_output_dict.values())
        assert len(torch_output) == len(
            onnx_output
        ), f"({len(torch_output)}, {len(onnx_output)})"

        for idx in range(len(onnx_output)):
            np.testing.assert_allclose(
                torch_output[idx].detach().cpu().numpy(),
                onnx_output[idx],
                rtol=1e-7,
                atol=1e-3,
            )

        print("accuracy test passed")



from .arithmetic import Arithmetic
from .batchnorm import BatchNorm
from .conv import Conv
from .embedding_layer import Embedding
from .Inspector_module import InspectorModule
from .layernorm import LayerNorm
from .linear import Linear
from .observer import Observer
from .pool import Pool
from .slice_layer import Slice
from .torch_layer import Arange, Full, Ones, Reshape, Size, Tensor



import os
import warnings

from .cli import convert
from .version import __version__


if os.path.dirname(os.path.realpath(__file__)) == os.path.join(
    os.path.realpath(os.getcwd()), "o2t"
):
    message = (
        "You are importing o2t within its own root folder ({}). "
        "This is not expected to work and may give errors. Please exit the "
        "o2t project source and relaunch your python interpreter."
    )
    warnings.warn(message.format(os.getcwd()))



import torch
import torch.nn as nn
from onnx_graphsurgeon import Constant


class Slice(nn.Module):
    @classmethod
    def from_onnx(cls, mod, env):
        def get_input_node(input, env, idx, default_value=-1):
            if isinstance(input, Constant):
                return input.values[idx]
            elif input is None:
                return default_value
            else:
                for feed in input.inputs:  # user is a Node
                    if (
                        feed.op == "Split"
                    ):  # for split node, we need to get the output of split node
                        return env[input.name]
                    else:
                        return env[feed.name]

        start = mod.inputs[1]
        end = mod.inputs[2]
        axes = mod.inputs[3] if len(mod.inputs) > 3 else None
        steps = mod.inputs[4] if len(mod.inputs) > 4 else None

        if max(axes.values) > 0:
            slice_inputs = (
                [slice(None, None, None)] * (max(axes.values) + 1)
                if axes is not None
                else [slice(None, None, None)] * len(start.values)
            )
            if axes is not None:
                for idx in range(len(axes.values)):
                    axes_idx = axes.values[idx]

                    start_ = get_input_node(start, env, idx)
                    end_ = get_input_node(end, env, idx)
                    steps_ = get_input_node(steps, env, idx, 1)

                    slice_inputs[axes_idx] = slice(start_, end_, steps_)
            else:
                raise NotImplementedError

            return tuple(slice_inputs)
        else:
            slice_inputs = [...] + [slice(None, None, None)] * (abs(max(axes.values)))
            for idx in range(len(axes.values)):
                axes_idx = axes.values[idx]

                start_ = get_input_node(start, env, idx)
                end_ = get_input_node(end, env, idx)
                steps_ = get_input_node(steps, env, idx, 1)

                slice_inputs[axes_idx] = slice(start_, end_, steps_)
            return tuple(slice_inputs)



import torch
import torch.nn as nn
from torch.nn.parameter import Parameter

from .utils import get_value_by_key


class Pool(nn.Module):
    @classmethod
    def from_onnx(cls, onnx_node):
        if onnx_node.op == "GlobalAveragePool":
            pool = nn.AdaptiveAvgPool2d((1, 1))
        elif onnx_node.op == "MaxPool":
            pool = nn.MaxPool2d(
                kernel_size=onnx_node.attrs["kernel_shape"],
                stride=onnx_node.attrs["strides"],
                padding=onnx_node.attrs["pads"][2:],
                ceil_mode=bool(get_value_by_key(onnx_node, "ceil_mode", 0)),
            )
        elif onnx_node.op == "AveragePool":
            pool = nn.AvgPool2d(
                kernel_size=onnx_node.attrs["kernel_shape"],
                stride=onnx_node.attrs["strides"],
                padding=get_value_by_key(onnx_node, "pads", [0, 0, 0, 0])[2:],
                ceil_mode=bool(get_value_by_key(onnx_node, "ceil_mode", 0)),
            )
        return pool



import torch
import torch.nn as nn
from torch.nn.parameter import Parameter


class Linear(nn.Module):
    @classmethod
    def from_onnx(cls, mod):
        if mod.attrs["transB"]:
            weight = Parameter(torch.from_numpy(mod.inputs[1].values))
        else:
            weight = Parameter(torch.from_numpy(mod.inputs[1].values).transpose(1, 0))
        bias = (
            None
            if len(mod.inputs) < 3
            else Parameter(torch.from_numpy(mod.inputs[2].values))
        )
        linear = nn.Linear(
            in_features=weight.shape[1],
            out_features=weight.shape[0],
            bias=bias is not None,
        )
        linear.weight = weight
        linear.bias = bias
        return linear



import torch
import torch.nn as nn


class BaseModule(nn.Module):
    def __init__(self):
        super().__init__()
        self.register_buffer("dummy_buffer", torch.Tensor(), persistent=False)


class Tensor(BaseModule):
    # Tensor module is needed, because torch.tensor is not supported in torch.fx
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return torch.tensor(x).to(self.dummy_buffer.device)

    @classmethod
    def from_onnx(cls):
        return cls()


class Size(BaseModule):
    # Size module is needed, because torch.Size is not supported in torch.fx
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return torch.Size(x)

    @classmethod
    def from_onnx(cls):
        return cls()


class Full(BaseModule):
    # Full module is needed, because torch.full is not supported in torch.fx
    def __init__(self, fill_value):
        super().__init__()
        self.fill_value = fill_value

    def forward(self, x):
        return torch.full(x, self.fill_value, device=self.dummy_buffer.device)

    @classmethod
    def from_onnx(cls, onnx_node):
        value = onnx_node.attrs.get("value")
        return cls(int(value.values))


class Reshape(nn.Module):
    # Reshape module is needed, because torch.reshape is not supported in torch.fx
    def __init__(self):
        super().__init__()

    def forward(self, x, shape):
        return torch.reshape(x, shape)

    @classmethod
    def from_onnx(cls):
        return cls()


class Arange(BaseModule):
    # Arange module is needed, because torch.reshape is not supported in torch.fx
    def __init__(self):
        super().__init__()

    def forward(self, start, end, step=1):
        return torch.arange(start, end, step, device=self.dummy_buffer.device)

    @classmethod
    def from_onnx(cls):
        return cls()


class Ones(BaseModule):
    # Ones module is needed, because torch.ones is not supported in torch.fx
    def __init__(self):
        super().__init__()

    def forward(self, size):
        return torch.ones(size, device=self.dummy_buffer.device)

    @classmethod
    def from_onnx(cls):
        return cls()



import torch
import torch.nn as nn


class Embedding(nn.Module):
    @classmethod
    def from_onnx(cls, mod):
        weight = nn.Parameter(
            torch.from_numpy(mod.inputs[0].values), requires_grad=False
        )
        num_embeddings, embedding_dim = weight.shape
        embedding = nn.Embedding(num_embeddings, embedding_dim)
        embedding.weight = weight

        return embedding



import torch
import torch.nn as nn
from torch.nn.parameter import Parameter


class Conv(nn.Module):
    @classmethod
    def from_onnx(cls, mod):
        weight = Parameter(torch.from_numpy(mod.inputs[1].values))
        weight_shape = weight.shape
        bias = (
            None
            if len(mod.inputs) < 3
            else Parameter(torch.from_numpy(mod.inputs[2].values))
        )
        if len(weight_shape) == 3:
            groups = mod.attrs["group"]
            padding = mod.attrs["pads"][0:1]
            conv = nn.Conv1d(
                in_channels=weight.shape[1] * groups,
                out_channels=weight.shape[0],
                kernel_size=weight.shape[2:],
                stride=mod.attrs["strides"],
                padding=padding,
                dilation=mod.attrs["dilations"],
                groups=groups,
                bias=bias is not None,
            )
            conv.weight = weight
            conv.bias = bias
        elif len(weight_shape) == 4:
            groups = mod.attrs["group"]
            padding = mod.attrs["pads"][0:2]
            conv = nn.Conv2d(
                in_channels=weight.shape[1] * groups,
                out_channels=weight.shape[0],
                kernel_size=weight.shape[2:],
                stride=mod.attrs["strides"],
                padding=padding,
                dilation=mod.attrs["dilations"],
                groups=groups,
                bias=bias is not None,
            )
            conv.weight = weight
            conv.bias = bias

        return conv



import torch
import torch.nn as nn
from torch.nn.parameter import Parameter


class Observer(nn.Module):
    def __init__(self, scale, zero_point, input_shape=None):
        super(Observer, self).__init__()
        self.scale = scale
        self.zero_point = zero_point
        self.input_shape = input_shape

    def calculate_qparams(self):
        return self.scale, self.zero_point

    def forward(self, x):
        return x



from setuptools import find_packages, setup

with open("VERSION", "r") as f:
    version = f.read().strip()

with open("o2t/version.py", "w") as f:
    f.write(f'__version__ = "{version}"\n')

setup(
    name="o2t",
    version=version,
    description="onnx2torch: onnx to native pytorch converter",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/inisis/onnx2torch",
    author="inisis",
    author_email="desmond.yao@buaa.edu.cn",
    project_urls={
        "Bug Tracker": "https://github.com/inisis/onnx2torch/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    license="MIT",
    install_requires=["loguru", "torch", "onnx"],
    packages=find_packages(exclude=("tests", "tests.*")),
    entry_points={"console_scripts": ["onnx2torch=o2t.cli:main"]},
    zip_safe=True,
    python_requires=">=3.6",
)



import torch
import torch.nn as nn
from torch.nn.parameter import Parameter


class LayerNorm(nn.Module):
    @classmethod
    def from_onnx(cls, mod):
        eps = mod.attrs.get("epsilon", 1e-5)
        if len(mod.inputs) == 1:
            normalized_shape = mod.inputs[0].shape
            return nn.LayerNorm(normalized_shape, eps, False, False)
        else:
            normalized_shape = mod.inputs[1].shape
            weight = Parameter(torch.from_numpy(mod.inputs[1].values))
            bias = (
                None
                if len(mod.inputs) < 3
                else Parameter(torch.from_numpy(mod.inputs[2].values))
            )

            layernorm = nn.LayerNorm(
                normalized_shape,
                eps=eps,
                elementwise_affine=True,
                bias=bias is not None,
            )
            layernorm.weight = weight
            layernorm.bias = bias

            return layernorm



from typing import Dict

import numpy as np

import onnx
import torch


def onnx_dtype_to_numpy(onnx_dtype: int) -> np.dtype:
    import onnx.mapping as mapping

    return np.dtype(mapping.TENSOR_TYPE_TO_NP_TYPE[onnx_dtype])


def gen_onnxruntime_input_data(model: onnx.ModelProto) -> Dict[str, np.array]:
    input_info = []
    for input_tensor in model.graph.input:
        name = input_tensor.name
        shape = []
        for dim in input_tensor.type.tensor_type.shape.dim:
            if dim.HasField("dim_param"):
                shape.append(dim.dim_param)
            elif dim.HasField("dim_value"):
                shape.append(dim.dim_value)
            else:
                shape.append(None)
        dtype = onnx_dtype_to_numpy(input_tensor.type.tensor_type.elem_type)

        input_info.append([name, shape, dtype])

    input_data_dict = {}
    for name, shapes, dtype in input_info:
        shapes = [
            shape if (shape != -1 and not isinstance(shape, str)) else 16
            for shape in shapes
        ]
        shapes = shapes if shapes else [1]
        if dtype in [np.int32, np.int64]:
            random_data = np.random.randint(10, size=shapes).astype(dtype)
        else:
            random_data = np.random.rand(*shapes).astype(dtype)
        input_data_dict[name] = random_data

    return input_data_dict


def onnxruntime_inference(
    model: onnx.ModelProto, input_data: dict
) -> Dict[str, np.array]:
    import onnxruntime as rt

    sess = rt.InferenceSession(
        model.SerializeToString(), providers=["CPUExecutionProvider"]
    )
    onnx_output = sess.run(None, input_data)

    output_names = [output.name for output in sess.get_outputs()]
    onnx_output = dict(zip(output_names, onnx_output))

    return onnx_output


numpy_to_torch_dtype_dict = {
    np.dtype(bool): torch.bool,
    np.dtype(np.uint8): torch.uint8,
    np.dtype(np.int8): torch.int8,
    np.dtype(np.int16): torch.int16,
    np.dtype(np.int32): torch.int32,
    np.dtype(np.int64): torch.int64,
    np.dtype(np.float16): torch.float16,
    np.dtype(np.float32): torch.float32,
    np.dtype(np.float64): torch.float64,
    np.dtype(np.complex64): torch.complex64,
    np.dtype(np.complex128): torch.complex128,
}


def numpy_dtype_to_torch(scalar_type):
    return numpy_to_torch_dtype_dict[scalar_type]



import torch


class InspectorModule(torch.nn.Module):
    def __init__(self):
        super(InspectorModule, self).__init__()

    def forward(self, *inputs):
        print(inputs)

        return inputs
